; ______________________________________________________________________________
;|^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
;|   MY-UTILS.LSP - utilities written in common lisp
;|______________________________________________________________________________
; ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
;

(defun printt (this)
    (format t "~a~%" this))

;______________________________________________________________________________
;

(defun princc (&rest params)
    (dolist (sym params)
        (if (eq sym t) (terpri) (princ sym))))

;______________________________________________________________________________
;
; store hash key with data - case insensitive

(defun put-data (key data dict)
    (setf (gethash (string-upcase key) dict)
        (cons key data)))

;______________________________________________________________________________
;
; get hash data from key - case insensitive

(defun get-data (key dict)
    (gethash (string-upcase key) dict))

;______________________________________________________________________________
;
;; (time (dotimes (i 100000) (join '("a" "b" "c") "-")))

(defun join (word-list separator)
    (apply #'concatenate 'string
        (princ-to-string (first word-list))
        (mapcan #'(lambda(x) (list separator (princ-to-string x)))
            (rest word-list))))

;______________________________________________________________________________
;

(defun join-all (wlist sepchr)
    (if wlist
        (reduce #'(lambda (x y)
              (concatenate 'string
                  (if (listp x) (join x sepchr) (princ-to-string x))
                  sepchr
                  (if (listp y) (join y sepchr) (princ-to-string y))))
            wlist)))

;____________________________________________________________________________
;

(defun begins (string sub)
    (if (eq (search sub string :test #'equalp) 0) t nil))

;____________________________________________________________________________
;

(defun skip-char (string pos sepchar max)
    (loop (if (>= (setq pos (1+ pos)) max)
              (return (1- pos))
              (if (char/= (char string pos) sepchar)
                  (return (1- pos))))))

;______________________________________________________________________________
;
; STRING-LIST creates a list of words by a string of chars.
; string   -> name of the string to transform in list.
; sepchar  -> the separation character.
; flag=t   -> it transforms multiple separators into one single.
; ritorna  -> the list of words that they compose the string.

(defun string-list (string sepchar flag)
    (let* ((listw nil)
           (lens 0) (oldp -1) (i 0))
        (if flag
            (setf string
                (string-trim (string sepchar) string)))
        (setf lens (length string)
              listw (if (eq lens 0) '("") nil)
              string (concatenate 'string string (string sepchar)))
        (if (> lens 0)
            (loop
                (when (char= (char string i) sepchar)
                    (setf listw (nconc listw (list (subseq string (1+ oldp) i))))
                    (if flag (setq i (skip-char string i sepchar lens)))
                    (setq oldp i))
                (if (> (setq i (1+ i)) lens)
                    (return))))
        listw))

;____________________________________________________________________________
;

(defun str-num (str)
    (let* ((lens (length str))
           (smx 0)
           (chr 0))
        (dotimes (i lens smx)
            (setq chr (char-code (char str i)))
            (if (or (> chr 57) (< chr 48))
                (return
                    (/ smx (expt 10 (- lens i))))
                (setq smx
                    (+ smx (* (expt 10 (- lens 1 i))
                              (- chr 48))))))))

;____________________________________________________________________________
;
; expand strings, replacing old string with new

(defun change-str (string old new)
    (let* ((from 0)
           (opos 0)
           (lnew (length new)))
        (loop
            (setq opos (search old string :start2 from))
            (if (not opos)
                (return string))
            (setf string (concatenate 'string (subseq string 0 opos) new
                                              (subseq string (1+ opos)))
                  from (+ opos lnew)))))

;____________________________________________________________________________
;

(defun fncount (fname)
   (let* ((wlist nil))
    (do* ((filep (open (addpath fname) :direction :input))
          (exline (read-line filep nil) (read-line filep nil)))
         ((null exline) (close filep)
            (progn
               (printt (length wlist))
             ;;(printt (length (sort wlist #'string-lessp)))
               (printt (length (remove-duplicates wlist :test #'equalp)))
               (sort wlist #'string-lessp)))
      (if (begins exline "(defun")
          (setf wlist (nconc wlist
                       (list (second (string-list exline #\space t)))))))))

;____________________________________________________________________________
;
; remove one-char atoms from clist except those in the chars string.

(defun remove-atom (clist chars)   ;warning! => (search "" "abc") => 0 !!!
    (delete-if #'(lambda (x) (search x chars)) clist))

;____________________________________________________________________________
;

(defun strpbrk (charray chars)
    (dotimes (i (length chars))
        (if (search (string (char chars i)) charray)
            (return t))))

;____________________________________________________________________________
;

(defun get-entries (from-nth to-nth wlist)
    (let* ((llst (length wlist)))
      (unless (and (> from-nth llst) (> to-nth llst))
         (unless (and (< from-nth 1) (< to-nth 1))
              (unless (> from-nth to-nth)
                (when (< from-nth 1) (setq from-nth 1))
                (when (> to-nth llst) (setq to-nth llst))
                (subseq wlist (1- from-nth) to-nth))))))

;____________________________________________________________________________
;
; number to string conversion

(defun num-str (num)
    (princ-to-string num))

;____________________________________________________________________________
;

(defun force-garbage (times)
    (dotimes (i times)
        (gc)))  ;force the garbage collection

;____________________________________________________________________________
;
; input sentence

(defun input (prompt clist)
    (let* ((str nil)
           (promptt prompt))
        (loop
            (clear-input)
            (if (not (equal str ""))
                (if clist (setf promptt (concatenate 'string prompt "("
                    (string-downcase (join clist "/")) "): "))))
            (if (not (and (eq *xlisp* 'corman) (equal str "")))
                (princ promptt))
            (setf str (read-line))
            (when (not (equal str ""))
                (if (eq *xlisp* 'corman)
                    (setf str (subseq str (length promptt))))
                (when (not (and clist (not (find str clist :test #'equalp))))
                    (setf str (string-trim " " str))
                    (return (if clist (string-downcase str) str)))))))

;____________________________________________________________________________
;
; output answer

(defun output2 (prompt this)
    (format t "~a~a~%~%" prompt this))

;____________________________________________________________________________
;
; carica un file testo in una lista con una struttura predefinita
; fname -> nome del file

(defun caricafile (fname dest)
    (let* ((myfile (open fname :direction :input))
           (count 0)
           (linea "")
           (blist nil))
        (loop
            (if (not (setf linea (read-line myfile nil)))
                (return))
            (if (eq (type-of dest) 'hash-table)
                (progn                               ;- hash table
                    (setf blist (read-from-string linea))
                    (put-data (first blist) (rest blist) dest))
                (progn                               ;- array
                    (setf blist (string-list linea #\~ nil)
                          (aref dest count) blist)
                    (setq count (1+ count)))))
        (close myfile)))

;____________________________________________________________________________
;

(defun savefile (hashtab fname prompt)
    (let* ((answ "s")
           (myfile nil)
           (*print-pretty* nil))
        (when prompt
            (princc "Sto per salvare \"" fname "\"." t)
            (setf answ (input "Desideri salvare la base di conoscenza? " '("s" "n"))))
        (when (equal answ "s")
            (setf myfile
                (open fname :direction :output :if-exists :supersede))
            (maphash #'(lambda (key val)
                (if prompt (printt key)) (format myfile "~s~%" val)) hashtab)
            (close myfile))))

