//-----------------------------------------------------------------
// XTRAWIN.C - Xlisp library that allow to call win32 api functions
//-----------------------------------------------------------------

#define entry(num,lst)   entrx(num,lst,seplist)     //extract an entry from a list of entries
#define max_entries(lst) max_entrixs(lst,seplist)   //count the max number of entries
#define VAL(a)           (double)atof(a)
#define vlimit           99

typedef UCHAR *PST;              //definition of new string pointer type
static  UCHAR seplist  = '\x7f'; //char separator of one entry within of an string-list
static  HWND  hwndx[vlimit+1];
static  INT   wincount = 0;
static  MSG   LastMsg[vlimit+1];
static  INT   LastMark = -1;
static  INT   UserMark = -1;

void    parseCmdLine  (LPSTR lpCmdLine, int *Xargc, char *Xargv[]);
void    CenterWin     (HWND hWndw);
HWND    DoMessage     (PST message);
PST     entrx         (int numero, PST lista, int sepchr);
int     max_entrixs   (PST lista, int sepchr);
BOOL    begins        (PST match1, PST match2);
BOOL    eequal        (PST match1, PST match2);
void    close_all_win (void);
void    optimize_hwnd (HWND from);
void    init_win32_api(HWND main_hwnd);

LRESULT WndProc2      (HWND hwndp, UINT message, WPARAM wParam, LPARAM lParam);


//----------------------------------------------------------------------
//----------------------------------------------------------------------
void init_win32_api(HWND main_hwnd) {

   hwndx[0] = main_hwnd;

}

//----------------------------------------------------------------------
// Split command line into arguments
//----------------------------------------------------------------------
void parseCmdLine(LPSTR lpCmdLine, int *Xargc, char *Xargv[]) {

     static int i, firstch;
     static unsigned char pszArgs[999];
     static unsigned char szModule[MAX_PATH+1];

     // initialise all
     strcpy(pszArgs, lpCmdLine);
     memset(szModule, '\0', sizeof(szModule));
     GetModuleFileName(hInst, szModule, MAX_PATH);

     /* first argument is always present */
     *Xargc   = 1;
     Xargv[0] = szModule;

     /* Split command line into arguments */
     // step through command line replacing spaces with zeros, initialise array
     i = 0;
     firstch = TRUE;
     while (pszArgs[i] != '\0') {
        if (isspace(pszArgs[i])) {
           pszArgs[i] = '\0';
           firstch = TRUE;
        }
        else if (firstch == TRUE) {
           Xargv[*Xargc] = &pszArgs[i];
           *Xargc += 1;
           firstch = FALSE;
        }
        i++;
     }
}

//----------------------------------------------------------------------
//----------------------------------------------------------------------
void CenterWin (HWND hWndw) {

     RECT rect;
     int width, height, xres, yres;

     GetWindowRect(hWndw, &rect);

     width  = rect.right  - rect.left;
     height = rect.bottom - rect.top;
     xres = GetSystemMetrics(SM_CXSCREEN);
     yres = GetSystemMetrics(SM_CYSCREEN);

     SetWindowPos(hWndw, HWND_TOP, (xres - width) / 2, \
                      (yres - height) / 2, 0, 0, SWP_NOSIZE);

}

//----------------------------------------------------------------------
// xcallwinapi - send command string to win32 api
//             - require one string argument
//----------------------------------------------------------------------
LVAL xcallwinapi()
{
     LVAL string_arg;
     string_arg = xlgastring();  /* "XLisp Get A STRING argument"   */
     xllastarg();  /* Issue an error message if there are any extra arguments */
     return cvfixnum((FIXTYPE)DoMessage(getstring(string_arg)));
}

//----------------------------------------------------------------------
// xstringptr  - return the pointer of the string that has been passed
//             - require one string argument
//----------------------------------------------------------------------
LVAL xstringptr()
{
     LVAL string_arg;
     string_arg = xlgastring();  /* "XLisp Get A STRING argument"   */
     xllastarg();  /* Issue an error message if there are any extra arguments */
     return cvfixnum((FIXTYPE)getstring(string_arg));
}

//----------------------------------------------------------------------
// xgetactivewin - return the active window that has the focus
//               - require no arguments
//----------------------------------------------------------------------
LVAL xgetactivewin()
{
     xllastarg();  /* Issue an error message if there are any argument */
     return cvfixnum((FIXTYPE)GetActiveWindow());
}

//----------------------------------------------------------------------
// xgetfocus     - return the active control that has the focus
//               - require no arguments
//----------------------------------------------------------------------
LVAL xgetfocus()
{
     xllastarg();  /* Issue an error message if there are any argument */
     return cvfixnum((FIXTYPE)GetFocus());
}

//----------------------------------------------------------------------
// xgetwinmessage - return the next msg values inserted into messages queue
//                - require no arguments
//----------------------------------------------------------------------
LVAL xgetwinmessage()
{
     static INT temp;
     LVAL result;

     temp=UserMark;
     if (++temp>vlimit) temp=0;

     if (temp != LastMark) {
        UserMark=temp;
        /* Cons it all together to produce a return value: */
        result = cons(cvfixnum((FIXTYPE)LastMsg[UserMark].hwnd),
                   cons(cvfixnum((FIXTYPE)LastMsg[UserMark].message),
                     cons(cvfixnum((FIXTYPE)LastMsg[UserMark].wParam),
                       cons(cvfixnum((FIXTYPE)LastMsg[UserMark].lParam),
                         NIL))));
     }else
        result = NIL;

     return result;
}

//----------------------------------------------------------------------
// xshowwindow   - show the window with normal/maximize/minimize/etc attribute
//               - require two arguments: window-handle and attribute
//----------------------------------------------------------------------
LVAL xshowwindow()
{
     LVAL integer1_arg;
     LVAL integer2_arg;
     integer1_arg = xlgafixnum();  /* "XLisp Get A FIXNUM argument"   */
     integer2_arg = xlgafixnum();  /* "XLisp Get A FIXNUM argument"   */
     xllastarg();  /* Issue an error message if there are any extra arguments */
     return cvfixnum((FIXTYPE)ShowWindow((HWND)getfixnum(integer1_arg),
                                              getfixnum(integer2_arg)));
}

//----------------------------------------------------------------------
// xcenterwindow - center the window on the desktop
//               - require one argument: window-handle
//----------------------------------------------------------------------
LVAL xcenterwindow()
{
     LVAL integer_arg;
     integer_arg = xlgafixnum();  /* "XLisp Get A FIXNUM argument"   */
     xllastarg();  /* Issue an error message if there are any extra arguments */
     CenterWin((HWND)getfixnum(integer_arg));
     return NIL;
}

//----------------------------------------------------------------------
// xcenterwindow - center the window on the desktop
//               - require one argument: window-handle
//----------------------------------------------------------------------
LVAL xsetwincolor()
{
     LVAL integer1_arg;
     LVAL integer2_arg;
     integer1_arg = xlgafixnum();  /* "XLisp Get A FIXNUM argument"   */
     integer2_arg = xlgafixnum();  /* "XLisp Get A FIXNUM argument"   */
     xllastarg();  /* Issue an error message if there are any extra arguments */

     DeleteObject ((HBRUSH)
       SetClassLong ((HWND)getfixnum(integer1_arg), GCL_HBRBACKGROUND,
         (LONG)CreateSolidBrush ((COLORREF)getfixnum(integer2_arg))));

     InvalidateRect((HWND)getfixnum(integer1_arg), NULL, TRUE) ;
     UpdateWindow((HWND)getfixnum(integer1_arg));

     return NIL;
}

//----------------------------------------------------------------------
//----------------------------------------------------------------------
LVAL xrgbcolor()
{
     LVAL integer1_arg;
     LVAL integer2_arg;
     LVAL integer3_arg;
     integer1_arg = xlgafixnum();  /* "XLisp Get A FIXNUM argument"   */
     integer2_arg = xlgafixnum();  /* "XLisp Get A FIXNUM argument"   */
     integer3_arg = xlgafixnum();  /* "XLisp Get A FIXNUM argument"   */
     xllastarg();  /* Issue an error message if there are any extra arguments */
     return cvfixnum((FIXTYPE)RGB((BYTE)getfixnum(integer1_arg),
                                  (BYTE)getfixnum(integer2_arg),
                                  (BYTE)getfixnum(integer3_arg)));
}

/* partial copy from header "win.h"
#define GetBValue(rgb) ((BYTE) ((rgb) >> 16))
#define GetGValue(rgb) ((BYTE) (((WORD) (rgb)) >> 8))
#define GetRValue(rgb) ((BYTE) (rgb))
#define RGB(r,g,b) ((DWORD)(((BYTE)(r)|((WORD)(g) << 8))|(((DWORD)(BYTE)(b)) << 16)))
#define HANDLE_WM_NOTIFY(h,wP,lP,fn) (fn)((h),(int)(wP),(NMHDR *)(lP))
#define HIBYTE(w) ((BYTE) (((WORD) (w) >> 8) & 0xFF))
#define HIWORD(l) ((WORD) (((DWORD) (l) >> 16) & 0xFFFF))
#define LOBYTE(w) (((BYTE) (w)) & 0xFF)
#define LOWORD(l) (((WORD) (l)) & 0xFFFF)
#define MAKELONG(a,b) ((LONG) (((WORD) (a)) | ((DWORD) ((WORD) (b))) << 16))
#define MAKEWORD(a,b) ((WORD) (((BYTE) (a)) | ((WORD) ((BYTE) (b))) << 8))
#define INDEXTOOVERLAYMASK(i) ((i) << 8)
#define INDEXTOSTATEIMAGEMASK(i) ((i) << 12)
#define MAKEINTATOM(i) (LPTSTR) ((DWORD) ((WORD) (i)))
#define MAKEINTRESOURCE(i) (LPTSTR) ((DWORD) ((WORD) (i)))
*/


//----------------------------------------------------------------------
/* DoMessage - process command string received from xcallwinapi()
               --------------------------------------------------------------
               require: first argument -> parent window (mandatory) may be nil
                       second argument -> win32 keyword (mandatory)
               --------------------------------------------------------------
               if second argument begins with "create":
                       second argument -> "createwindow/button.." (mandatory)
                        third argument -> window caption (mandatory)
                       fourth argument -> window style   (optional) may be nil
                        fifth argument -> window pos.X   (optional)
                        sixth argument -> window pos.Y   (optional)
                      seventh argument -> window width   (optional)
                       eighth argument -> window height  (optional)
               else
                       second argument -> WIN32  message (mandatory)
                        third argument -> WPARAM message (optional)
                       fourth argument -> LPARAM message (optional) */
//----------------------------------------------------------------------
HWND DoMessage (PST message) {

     static int tipo;
     static UINT xstile, wpar, lpar;
     static HWND hwndret, hparent;
     static char xcommand[999];
     static char xcaption[999];
     static char xappname[999];
     static int xwin, ywin, xwid, yhei, maxpar;

     hwndret = NULL;
     hparent = (HWND)(UINT)VAL((char*)entry(1,message));
     maxpar  = max_entries(message);
     strcpy(xcommand,(char*)entry(2,message));

     if (begins((PST)xcommand,(PST)"Create")) {

        if (eequal((PST)xcommand,(PST)"CreateWindow") && wincount<vlimit) {
           wincount++;
           tipo = 1; strcpy(xappname,szAppName);
           xstile = WS_OVERLAPPEDWINDOW|WS_CLIPSIBLINGS;       //default window style
           xwin = CW_USEDEFAULT; ywin=xwin; xwid=400; yhei=300;  //default
        }
        else if (eequal((PST)xcommand,(PST)"CreateButton")) {
           tipo = 2; strcpy(xappname,"BUTTON");
           xstile = WS_CHILD|WS_VISIBLE|BS_PUSHBUTTON|WS_TABSTOP|WS_GROUP|WS_CLIPSIBLINGS; //default
           xwin=0; ywin=21; xwid=200; yhei=22;   //default
        }
        else if (eequal((PST)xcommand,(PST)"CreateEdit")) {
           tipo = 2; strcpy(xappname,"EDIT");
           xstile = WS_CHILD|WS_VISIBLE|WS_BORDER|WS_TABSTOP|WS_GROUP|WS_CLIPSIBLINGS; //default
           xwin=0; ywin=0; xwid=200; yhei=20;  //default
        }
        else if (eequal((PST)xcommand,(PST)"CreateComboBox")) {
           tipo = 2; strcpy(xappname,"COMBOBOX");
           xstile = CBS_DROPDOWN|WS_CHILD|WS_VISIBLE|WS_TABSTOP|WS_GROUP|WS_CLIPSIBLINGS; //default
           xwin=0; ywin=44; xwid=200; yhei=80;  //default
        }
        else if (eequal((PST)xcommand,(PST)"CreateListBox")) {
           tipo = 2; strcpy(xappname,"LISTBOX");
           xstile = WS_VSCROLL|LBS_DISABLENOSCROLL|WS_CHILD|WS_VISIBLE|WS_BORDER|WS_TABSTOP|WS_GROUP|WS_CLIPSIBLINGS; //default
           xwin=0; ywin=70; xwid=200; yhei=80;  //default
        }
        else if (eequal((PST)xcommand,(PST)"CreateMDIClient")) {  //IT DOESN'T WORK !!
           tipo = 2; strcpy(xappname,"MDICLIENT");
           xstile = WS_VSCROLL|WS_HSCROLL|WS_CLIPCHILDREN|WS_CHILD|WS_VISIBLE|WS_BORDER|WS_TABSTOP|WS_GROUP|WS_CLIPSIBLINGS; //default
           xwin=0; ywin=70; xwid=220; yhei=220;  //default
        }
        else if (eequal((PST)xcommand,(PST)"CreateScrollBar")) {
           tipo = 2; strcpy(xappname,"SCROLLBAR");
           xstile = SBS_VERT|WS_CHILD|WS_VISIBLE|WS_TABSTOP|WS_GROUP|WS_CLIPSIBLINGS; //default
           xwin=205; ywin=0; xwid=16; yhei=160;  //default
        }
        else if (eequal((PST)xcommand,(PST)"CreateStatic")) {
           tipo = 2; strcpy(xappname,"STATIC");
           xstile = SS_LEFTNOWORDWRAP|WS_CHILD|WS_VISIBLE|WS_CLIPSIBLINGS; //default
           xwin=0; ywin=140; xwid=200; yhei=20;  //default
        }
        else return hwndret;

                         //CAPTION
        strcpy(xcaption,(char*)entry(3,message));

        if (maxpar>=4)   //STYLE
           if ((UINT)VAL((char*)entry(4,message))>0)  //STYLE MUST BE GREATER THAN ZERO?
              xstile=(UINT)VAL((char*)entry(4,message));

        if (maxpar>=5)   //POS.X
           if (!eequal((PST)entry(5,message),(PST)"nil"))
              xwin=(INT)VAL((char*)entry(5,message));

        if (maxpar>=6)   //POS.Y
           if (!eequal((PST)entry(6,message),(PST)"nil"))
              ywin=(INT)VAL((char*)entry(6,message));

                         //WIDTH
        if (maxpar>=7) xwid=(INT)VAL((char*)entry(7,message));

                         //HEIGHT
        if (maxpar>=8) yhei=(INT)VAL((char*)entry(8,message));

                         //CREATE WINDOW
        hwndret = CreateWindow (xappname,             // window class name
                          xcaption,                   // window caption
                          xstile,                     // window style
                          xwin,                       // initial x position
                          ywin,                       // initial y position
                          xwid,                       // initial x size
                          yhei,                       // initial y size
                          hparent,                    // parent window handle
                          NULL,                       // window menu handle
                          hInst,                      // program instance handle
                          NULL) ;                     // creation parameters

        if (tipo == 1) {
           SetMenu(hwndret, NULL);   // no menu for now, sorry..
           if ((maxpar>=5&&xwin==CW_USEDEFAULT)||(maxpar>=6&&ywin==CW_USEDEFAULT))
              CenterWin(hwndret);
           hwndx[wincount] = hwndret;
           ShowWindow (hwndret, SW_SHOWNORMAL);
           UpdateWindow (hwndret) ;
        }

     } else {
        wpar=(WPARAM)0;   //default
        lpar=(LPARAM)0;   //default
        xstile=(UINT)VAL((char*)entry(2,message));
        if (maxpar>=3) wpar=(WPARAM)VAL((char*)entry(3,message));
        if (maxpar>=4) lpar=(LPARAM)VAL((char*)entry(4,message));
        hwndret = (HWND)SendMessage(hparent, xstile, wpar, lpar);
     }

     return hwndret;

}

//----------------------------------------------------------
// ENTRY: return an entry contained into string 'lista'
//----------------------------------------------------------
PST entrx (int numero, PST lista, int sepchr) {

  // ritona SEPCHR= quando l'elemento richiesto e' fuori dalla lista
  // "    ELEMENTO= quando l'elemento richiesto esiste

  static int limite;
  static PST posiz, oldpos;
  static UCHAR PT_boutput[999];

          //tramite swapbuf preparo il buffer di ritorno
  PT_boutput[0] = sepchr;        //mette sul buffer di ritorno il codice di errore di default
  PT_boutput[1] = '\0';          //end string

  if (numero && lista) {
     limite = 1;
     oldpos = lista;
     for (posiz = oldpos; *posiz; posiz++) {
        if (*posiz == sepchr) {
           if (++limite > numero) break;
           oldpos = posiz + 1;
        }}
     if (numero <= limite) {
        limite=0;
        for ( ; oldpos < posiz; oldpos++)
            PT_boutput[limite++] = *oldpos;
        PT_boutput[limite] = '\0';
     }
  }

  return PT_boutput;  //returns the pointer to local buffer

}

//-----------------------------------------------------------------------------
// EEQUAL: compare if two strings are perfectly equals (case insensitive)
//-----------------------------------------------------------------------------
BOOL eequal(PST match1, PST match2) {

  static int bb1;
  static int bb2;
  static int diff = 'A' - 'a';

  if (match1 && match2) {
     do {
       bb1=*match1++;
       bb2=*match2++;
       if (bb1>='a' && bb1 <= 'z') bb1+=diff;
       if (bb2>='a' && bb2 <= 'z') bb2+=diff;
       if (bb1 != bb2) return FALSE;
     } while (bb2);     //exit from here only when: bb1=0 and bb2=0
     return TRUE;
  }

  return FALSE;

}

//-----------------------------------------------------------------------------
// BEGINS: check if the string 'match1' begins with 'match2' (case insensitive)
//-----------------------------------------------------------------------------
BOOL begins(PST match1, PST match2) {

    static int bb1;
    static int bb2;
    static int diff = 'A' - 'a';

    if (match1 && match2) {
       do {
         bb1=*match1++;
         bb2=*match2++;
         if (bb1>='a' && bb1 <= 'z') bb1+=diff;
         if (bb2>='a' && bb2 <= 'z') bb2+=diff;
         if (bb1 != bb2 && bb2) return FALSE;
       } while (bb2);
       return TRUE;       //attention: begins("what","") returns TRUE !
    }

    return FALSE;

}

//----------------------------------------------------------------------
// MAX_ENTRIES: count the max number of entry inside the string 'lista'
//----------------------------------------------------------------------
int max_entrixs (PST lista, int sepchr) {

    static int limite;

    limite = 0;
    if (lista) {
       if (*lista) limite++;
       while (*lista) if (*lista++ == sepchr) limite++;
    }

    return limite;

}

//----------------------------------------------------------------------
// WndProc2 procedure is called by main window procedure 'WndProc()'
//----------------------------------------------------------------------
LRESULT WndProc2 (HWND hwndp, UINT message, WPARAM wParam, LPARAM lParam) {

     static HDC         hdc ;
     static PAINTSTRUCT ps ;
     //static RECT        rect ;

     if (++LastMark>vlimit) LastMark=0;
     LastMsg[LastMark].hwnd    = hwndp;
     LastMsg[LastMark].message = message;
     LastMsg[LastMark].wParam  = wParam;
     LastMsg[LastMark].lParam  = lParam;
     if (LastMark==UserMark)
        if (++UserMark>vlimit) UserMark=0;

     switch (message) {

       case WM_PAINT:
            //if (hwndp==GetActiveWindow()) {
            hdc = BeginPaint (hwndp, &ps) ;
            //GetClientRect (hwndp, &rect) ;
            //DrawText (hdc, TEXT ("Hello Windows!"), -1, &rect,
            //          DT_SINGLELINE | DT_CENTER | DT_VCENTER) ;
            EndPaint (hwndp, &ps) ;
            return 0 ;

       case WM_CLOSE:
            //MessageBox(NULL, "pooo", "ppp", MB_OK);
            optimize_hwnd(hwndp);  //array optimizazion
            break;

/*
       case WM_DESTROY:
            return 0 ;
*/
     }

     return DefWindowProc (hwndp, message, wParam, lParam);

}

//----------------------------------------------------------------------
// closes all windows at the exit from Xlisp
//----------------------------------------------------------------------
void close_all_win(void) {

     for (int ii = 1; ii<=wincount; ii++)
         SendMessage(hwndx[ii], WM_CLOSE, 0, 0);
     wincount = 0;

}

//----------------------------------------------------------------------
// array optimizazion
//----------------------------------------------------------------------
void optimize_hwnd(HWND from) {

     for (int ii = 1; ii<=wincount; ii++)
         if (hwndx[ii]==from) {
            hwndx[ii]=hwndx[wincount];  //replace closed handle with last
            wincount--;
            break;
         }
}
